/* mod_auth_pwcheck.c
 * Anton Voronin (anton@urc.ac.ru)
 *
 * It authenticates user via pwcheck daemon from Cyrus SASL package so that
 * httpd doesn't need to be run as root. The code for pwcheck interaction
 * is taken from imap/login_unix_pwcheck.c and lib/retry.c of Cyrus SASL
 * distribution archive.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <unistd.h>
#include <grp.h>

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#define MAX_RESPONSE 1024
#define SOCK_PATH "/var/pwcheck/pwcheck"


/*-----------------------------------------------------------------------*/

/* internal configuration data */
typedef struct {
	int auth_pwcheck_enabled;	/* contains true/false */
	int auth_pwcheck_authoritative;	/* contains true/false */
	int use_system_group_file;	/* contains true/false */
} auth_pwcheck_config_rec;

static void* auth_pwcheck_create_config( pool*, char* );
static const char* auth_pwcheck_command_handler( cmd_parms*, auth_pwcheck_config_rec*, int );
static int auth_pwcheck_authenticate( request_rec* );
int retry_writev( int, struct iovec*, int );
static int auth_pwcheck_authorize( request_rec* );


/*-----------------------------------------------------------------------*/

/* config commands */
static command_rec auth_pwcheck_config_cmds[] = {
	{
		"AuthPwcheck",
		auth_pwcheck_command_handler,
		NULL,
		OR_AUTHCFG,
		FLAG, 
		"authenticate user using pwcheck daemon"
	},
	{
		"AuthPwcheckAuthoritative",
		auth_pwcheck_command_handler,
		NULL,
		OR_AUTHCFG,
		FLAG, 
		"If set to Off, access control is passed to subsequent auth modules "
		"when mod_auth_pwcheck cannot authenticate or authorize the user."
	},
	{
		"UseSystemGroupFile",
		auth_pwcheck_command_handler,
		NULL,
		OR_AUTHCFG,
		FLAG, 
		"whether to use system group file instead of AuthGroupFile"
	},
	{
		NULL
	}
};

/* module */
module auth_pwcheck_module = {
	STANDARD_MODULE_STUFF,
	NULL,				/* module initializer */
	auth_pwcheck_create_config,	/* dir config creator */
	NULL,				/* dir config merger */
	NULL,				/* server config creator */
	NULL,				/* server config merger */
	auth_pwcheck_config_cmds,	/* config command table */
	NULL,				/* response handlers table */
	NULL,				/* filename-to-URI translation */
	auth_pwcheck_authenticate,	/* check user id */
	auth_pwcheck_authorize,		/* authorize user */
	NULL,				/* check host access */
	NULL,				/* check/set MIME type */
	NULL,				/* fixups */
	NULL,				/* logger */
	NULL,				/* header parser */
	NULL,				/* process initializer */
	NULL,				/* process exit/cleanup */
	NULL				/* post read_request handling */
};


/* config creator */
static void *auth_pwcheck_create_config( pool *p, char *dir_name )
{
	auth_pwcheck_config_rec *new_conf = ap_pcalloc( p, sizeof( auth_pwcheck_config_rec ) );
	new_conf->auth_pwcheck_authoritative = 1;
	new_conf->auth_pwcheck_enabled = 0;
	new_conf->use_system_group_file = 0;
	return new_conf;
}


/* config command handler */
static const char *auth_pwcheck_command_handler( cmd_parms *cmd, auth_pwcheck_config_rec *struct_ptr, int arg )
{
	if( strcmp( cmd->cmd->name, "AuthPwcheck" ) == 0 ){
		struct_ptr->auth_pwcheck_enabled = arg;
	}else if( strcmp( cmd->cmd->name, "AuthPwcheckAuthoritative" ) == 0 ){
		struct_ptr->auth_pwcheck_authoritative = arg;
	}else{
		struct_ptr->use_system_group_file = arg;
	}
	return NULL;
}


/* Authenticates user (checks ID and password).
Returns:
- OK if authentication was successful,
- AUTH_REQUIRED, if authentication failed, 
- SERVER_ERROR, if authentication is not possible,
- DECLINED if the module is not turned on via the configuration. */
static int auth_pwcheck_authenticate( request_rec *r )
{
	auth_pwcheck_config_rec *dir_config;	// our per-dir configuration
	conn_rec *c = r->connection;		// connection data
	const char *sent_pw;			// user-supplied password
	int res;				// ap_get_basic_auth return code
	int s;					// socket descriptor
	struct sockaddr_un srvaddr;		// socket address (in the communication space)
	struct iovec iov[ 10 ];			// write buffers array
	static char response[ MAX_RESPONSE ];	// pwcheck reply
	int start, n;				// counters

	// obtaining password supplied by user:
	if( ( res = ap_get_basic_auth_pw( r, &sent_pw ) ) != OK ){
		return res;
	}

	dir_config = ( auth_pwcheck_config_rec * )ap_get_module_config( r->per_dir_config, &auth_pwcheck_module );
	if( dir_config->auth_pwcheck_enabled == 0 ){
		// AuthPwcheck is not turned on in this directory
		return DECLINED;
	}

	// setting up unix socket connection to pwcheck:
	errno = 0;	/* clear errno just in case */
	s = socket(AF_UNIX, SOCK_STREAM, 0);
	if( s == -1 ){
		ap_log_rerror( APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
			"can't open socket for pwcheck server: %s", strerror( errno ) );
		ap_note_basic_auth_failure( r );
		return SERVER_ERROR;
	}
	memset( ( char * )&srvaddr, 0, sizeof( srvaddr ) );
	srvaddr.sun_family = AF_UNIX;
	strcpy( srvaddr.sun_path, SOCK_PATH );
	if( connect( s, ( struct sockaddr * )&srvaddr, sizeof( srvaddr ) ) == -1 ) {
		ap_log_rerror( APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
			"can't connect to pwcheck server: %s", strerror( errno ) );
		ap_note_basic_auth_failure( r );
		return SERVER_ERROR;
	}

	// initializing data to pass to pwcheck:
	iov[0].iov_base = c->user;
	iov[0].iov_len = strlen( c->user ) + 1;
	iov[1].iov_base = ( char * )sent_pw;
	iov[1].iov_len = strlen( sent_pw ) + 1;

	// passing data to pwcheck:
	retry_writev( s, iov, 2 );

	// reading pwcheck reply:
	start = 0;
	while( start < sizeof( response ) - 1 ){
		n = read( s, response + start, sizeof( response ) - 1 - start );
		if( n < 1 ) break;
		start += n;
	}
	close( s );
	response[ start ] = '\0';

	if( strcmp( response, "OK" ) == 0 ){
		// authentication succeeded
	}else if( dir_config->auth_pwcheck_authoritative ){
		// authentication failed
		ap_log_rerror( APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
			"auth for user %s failed: %s", c->user, response );
		ap_note_basic_auth_failure( r );
		return AUTH_REQUIRED;
	}else{
		return DECLINED;
	}
}


/* Keep calling the writev() system call until all the data is written out
or an error occurs.  */
int retry_writev(int fd, struct iovec *iov, int iovcnt) {

	int i, n;
	int written = 0;
    
	for (;;) {
		while (iovcnt && iov[0].iov_len == 0) {
			iov++;
			iovcnt--;
		}

		if (!iovcnt) return written;

		n = writev(fd, iov, iovcnt);
		if (n == -1) {
			if (errno == EINTR) continue;
			return -1;
		}

		written += n;

		for (i = 0; i < iovcnt; i++) {
			if (iov[i].iov_len > n) {
				iov[i].iov_base = (char *)iov[i].iov_base + n;
				iov[i].iov_len -= n;
				break;
			}
			n -= iov[i].iov_len;
			iov[i].iov_len = 0;
		}

		if (i == iovcnt) return written;
	}
}


/* Authorizes user (checks is the user allowed to access) in case if a system
group is required.
Returns:
- DECLINED if "required" keyword contains something else than system group,
- OK if the user belongs to the required system group,
- AUTH_REQUIRED, if the user is not authorized to access. */
static int auth_pwcheck_authorize( request_rec *r ){

	int i, j;			// loop variables
	const array_header *reqs_array;	// array of auth requirements
	require_line *reqs_array_elem;	// reqs_array element
	int method_restricted = 0;	// flag: 1 if there's at least one
					// requirement for the given method
	struct group *real_group;	// group structure for the required group
	const char *line, *word;		// a text line and a word within it
	auth_pwcheck_config_rec *dir_config;	// our per-dir configuration

	dir_config = ( auth_pwcheck_config_rec* )ap_get_module_config( r->per_dir_config, &auth_pwcheck_module );
	if( dir_config->use_system_group_file == 0 ){
		// system group authorization is not required,
		// we leave other authorization types to
		// corresponding modules.
       	        return DECLINED;
	}

	reqs_array = ap_requires( r );
	if( reqs_array == NULL ){
		return (OK);
	}
	reqs_array_elem = ( require_line* )reqs_array->elts;

	for( i = 0; i < reqs_array->nelts; i++ ){

		line = reqs_array_elem[ i ].requirement;
		word = ap_getword_white( r->pool, &line );
		if( strcmp( word, "group" ) == 0 ){
			// searching user in all equired groups:
			while( line[ 0 ] != '\0' ){
				word = ap_getword_conf( r->pool, &line );
        	        	real_group = getgrnam( word );
				if( real_group == NULL ){
					// required system group does not exist
					continue;
				}else{
					// checking user against a group:
					for( j = 0; real_group->gr_mem[ j ] != NULL; j++ ){
						if( strcmp( real_group->gr_mem[ j ], r->connection->user ) == 0 ){
							// authorization succeeded
							return OK;
						}
					}
				}
            		}
		}else if( strcmp( word, "user" ) ){
			while( line[ 0 ] != '\0' ){
				word = ap_getword_conf( r->pool, &line );
				if( strcmp( word, r->connection->user ) == 0 ){
					return OK;
				}
			}
		}else if( strcmp( word, "valid-user" ) ){
			return OK;
		}else if( dir_config->auth_pwcheck_authoritative ){
			ap_log_rerror( APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
				"%s: module mod_auth_pwcheck does not support \"%s\" require directive.",
				r->uri, word );
		}
	}

	if( dir_config->auth_pwcheck_authoritative ){
		// there was at least one requirement but it failed
		ap_note_basic_auth_failure( r );
		return AUTH_REQUIRED;
	}else{
		return DECLINED;
	}
}

